@@ -4,6 +4,8 @@ Fabric's own fabfile.
 
 from __future__ import with_statement
 
+import nose
+
 from fabric.api import *
 from fabric.contrib.project import rsync_project
 # Need to import this as fabric.version for reload() purposes
@@ -15,7 +17,7 @@ _version = fabric.version.get_version
 docs_host = 'jforcier@fabfile.org'
 
 
-def test(args=""):
+def test(args=None):
     """
     Run all unit tests and doctests.
 
@@ -23,7 +25,10 @@ def test(args=""):
     """
     default_args = "-sv --with-doctest --nologcapture --with-color"
     default_args += (" " + args) if args else ""
-    local('nosetests %s' % default_args, capture=False)
+    try:
+        nose.core.run(argv=[''] + default_args.split())
+    except SystemExit:
+        abort("Nose encountered an error; you may be missing newly added test dependencies. Try running 'pip install -r requirements.txt'.")
 
 
 def build_docs(clean='no', browse='no'):
@@ -36,18 +41,17 @@ def build_docs(clean='no', browse='no'):
     b = ""
     if browse.lower() in ['yes', 'y']:
         b = " && open _build/html/index.html"
-    local('cd docs; make %shtml%s' % (c, b), capture=False)
+    local('cd docs; make %shtml%s' % (c, b))
 
 
 @hosts(docs_host)
 def push_docs():
     """
-    Build and push the Sphinx docs to docs.fabfile.org
+    Build docs and zip for upload to RTD
     """
     build_docs(clean='yes')
-    remote_loc = '/var/www/docs.fabfile/%s/' % _version('short')
-    run('mkdir -p %s' % remote_loc)
-    rsync_project(remote_loc, 'docs/_build/html/', delete=True)
+    v = _version('short')
+    local("cd docs/_build/html && zip -r ../%s.zip ." % v)
 
 
 def _code_version_is_tagged():
@@ -62,15 +66,15 @@ def _update_code_version(force):
     """
     version_file = "fabric/version.py"
     raw_input("Work has been done since last tag, version update is needed. Hit Enter to load version info in your editor: ")
-    local("$EDITOR %s" % version_file, capture=False)
+    local("$EDITOR %s" % version_file)
     # Try to detect whether user bailed out of the edit
     if not local("git diff -- %s" % version_file) and not force:
         abort("You seem to have aborted the file edit, so I'm aborting too.")
     # Reload version module to get new version
     reload(fabric.version)
     # Commit the version update
-    local("git add %s" % version_file, capture=False)
-    local("git commit -m \"Cut %s\"" % _version('verbose'), capture=False)
+    local("git add %s" % version_file)
+    local("git commit -m \"Cut %s\"" % _version('verbose'))
 
 def _commits_since_tag():
     """
@@ -114,24 +118,24 @@ def tag(force='no', push='no'):
             f,
             _version('verbose'),
             _version('short')
-        ), capture=False)
+        ))
         # And push to the central server, if we were told to
         if push.lower() in ['y', 'yes']:
-            local("git push origin %s" % _version('short'), capture=False)
+            local("git push origin %s" % _version('short'))
 
 
 def build():
     """
     Build (but don't upload) via setup.py
     """
-    local('python setup.py sdist', capture=False)
+    local('python setup.py sdist')
 
 
 def upload():
     """
     Build, register and upload to PyPI
     """
-    local('python setup.py sdist register upload', capture=False)
+    local('python setup.py sdist register upload')
 
 
 def release(force='no'):
@@ -140,5 +144,3 @@ def release(force='no'):
     """
     tag(force=force, push='yes')
     upload()
-    with settings(host_string=docs_host):
-        push_docs()
