@@ -8,8 +8,13 @@ Context managers for use with the ``with`` statement.
 """
 
 from contextlib import contextmanager, nested
+import sys
 
-from fabric.state import env, output
+from fabric.state import env, output, win32
+
+if not win32:
+    import termios
+    import tty
 
 
 def _set_output(groups, which):
@@ -81,7 +86,8 @@ def _setenv(**kwargs):
     """
     previous = {}
     for key, value in kwargs.iteritems():
-        previous[key] = env[key]
+        if key in env:
+            previous[key] = env[key]
         env[key] = value
     try:
         yield
@@ -139,14 +145,14 @@ def cd(path):
     """
     Context manager that keeps directory state when calling operations.
 
-    Any calls to `run`, `sudo` or `local` within the wrapped block will
+    Any calls to `run`, `sudo`, `get`, or `put` within the wrapped block will
     implicitly have a string similar to ``"cd <path> && "`` prefixed in order
-    to give the sense that there is actually statefulness involved.
+    to give the sense that there is actually statefulness involved.  `cd` only
+    affects the remote paths for `get` and `put` -- local paths are untouched.
 
     Because use of `cd` affects all such invocations, any code making use of
-    `run`/`sudo`/`local`, such as much of the ``contrib`` section, will also be
-    affected by use of `cd`. However, at this time, `get` and `put` do not
-    honor `cd`; we expect this to be addressed in future releases.
+    those operations, such as much of the ``contrib`` section, will also be
+    affected by use of `cd`.
 
     Like the actual 'cd' shell builtin, `cd` may be called with relative paths
     (keep in mind that your default starting directory is your remote user's
@@ -180,9 +186,141 @@ def cd(path):
         future, so we do not recommend manually altering ``env.cwd`` -- only
         the *behavior* of `cd` will have any guarantee of backwards
         compatibility.
+
+    .. note::
+
+        Space characters will be escaped automatically to make dealing with
+        such directory names easier.
+
+    .. versionchanged:: 1.0
+        Applies to `get` and `put` in addition to the command-running
+        operations.
     """
-    if env.get('cwd') and not path.startswith('/'):
-        new_cwd = env.cwd + '/' + path
+    return _change_cwd('cwd', path)
+
+
+def lcd(path):
+    """
+    Context manager for updating local current working directory.
+
+    This context manager is identical to `~fabric.context_managers.cd`, except
+    that it changes a different env var (`lcwd`, instead of `cwd`) and thus
+    only affects the invocation of `~fabric.operations.local` and the local
+    arguments to `~fabric.operations.get`/`~fabric.operations.put`.
+
+    .. versionadded:: 1.0
+    """
+    return _change_cwd('lcwd', path)
+
+
+def _change_cwd(which, path):
+    path = path.replace(' ', '\ ')
+    if env.get(which) and not path.startswith('/'):
+        new_cwd = env.get(which) + '/' + path
     else:
         new_cwd = path
-    return _setenv(cwd=new_cwd)
+    return _setenv(**{which: new_cwd})
+
+
+def path(path, behavior='append'):
+    """
+    Append the given ``path`` to the PATH used to execute any wrapped commands.
+
+    Any calls to `run` or `sudo` within the wrapped block will implicitly have
+    a string similar to ``"PATH=$PATH:<path> "`` prepended before the given
+    command.
+
+    You may customize the behavior of `path` by specifying the optional
+    ``behavior`` keyword argument, as follows:
+
+    * ``'append'``: append given path to the current ``$PATH``, e.g.
+      ``PATH=$PATH:<path>``. This is the default behavior.
+    * ``'prepend'``: prepend given path to the current ``$PATH``, e.g.
+      ``PATH=<path>:$PATH``.
+    * ``'replace'``: ignore previous value of ``$PATH`` altogether, e.g.
+      ``PATH=<path>``.
+
+    .. note::
+
+        This context manager is currently implemented by modifying (and, as
+        always, restoring afterwards) the current value of environment
+        variables, ``env.path`` and ``env.path_behavior``. However, this
+        implementation may change in the future, so we do not recommend
+        manually altering them directly.
+
+    .. versionadded:: 1.0
+    """
+    return _setenv(path=path, path_behavior=behavior)
+
+
+def prefix(command):
+    """
+    Prefix all wrapped `run`/`sudo` commands with given command plus ``&&``.
+
+    This is nearly identical to `~fabric.operations.cd`, except that nested
+    invocations append to a list of command strings instead of modifying a
+    single string.
+
+    Most of the time, you'll want to be using this alongside a shell script
+    which alters shell state, such as ones which export or alter shell
+    environment variables.
+
+    For example, one of the most common uses of this tool is with the
+    ``workon`` command from `virtualenvwrapper
+    <http://www.doughellmann.com/projects/virtualenvwrapper/>`_::
+
+        with prefix('workon myvenv'):
+            run('./manage.py syncdb')
+
+    In the above snippet, the actual shell command run would be this::
+
+        $ workon myvenv && ./manage.py syncdb
+
+    This context manager is compatible with `~fabric.context_managers.cd`, so
+    if your virtualenv doesn't ``cd`` in its ``postactivate`` script, you could
+    do the following::
+
+        with cd('/path/to/app'):
+            with prefix('workon myvenv'):
+                run('./manage.py syncdb')
+                run('./manage.py loaddata myfixture')
+
+    Which would result in executions like so::
+
+        $ cd /path/to/app && workon myvenv && ./manage.py syncdb
+        $ cd /path/to/app && workon myvenv && ./manage.py loaddata myfixture
+
+    Finally, as alluded to near the beginning,
+    `~fabric.context_managers.prefix` may be nested if desired, e.g.::
+
+        with prefix('workon myenv'):
+            run('ls')
+            with prefix('source /some/script'):
+                run('touch a_file')
+
+    The result::
+
+        $ workon myenv && ls
+        $ workon myenv && source /some/script && touch a_file
+
+    Contrived, but hopefully illustrative.
+    """
+    return _setenv(command_prefixes=env.command_prefixes + [command])
+
+
+@contextmanager
+def char_buffered(pipe):
+    """
+    Force local terminal ``pipe`` be character, not line, buffered.
+
+    Only applies on Unix-based systems; on Windows this is a no-op.
+    """
+    if win32 or not sys.stdin.isatty():
+        yield
+    else:
+        old_settings = termios.tcgetattr(pipe)
+        tty.setcbreak(pipe)
+        try:
+            yield
+        finally:
+            termios.tcsetattr(pipe, termios.TCSADRAIN, old_settings)
