@@ -14,9 +14,9 @@ from optparse import OptionParser
 import os
 import sys
 
-from fabric import api # For checking callables against the API 
+from fabric import api # For checking callables against the API
 from fabric.contrib import console, files, project # Ditto
-from fabric.network import denormalize, normalize
+from fabric.network import denormalize, interpret_host_string, disconnect_all
 from fabric import state # For easily-mockable access to roles, env and etc
 from fabric.state import commands, connections, env_options
 from fabric.utils import abort, indent
@@ -45,28 +45,48 @@ def load_settings(path):
     return {}
 
 
+def _is_package(path):
+    """
+    Is the given path a Python package?
+    """
+    return (
+        os.path.isdir(path)
+        and os.path.exists(os.path.join(path, '__init__.py'))
+    )
+
+
 def find_fabfile():
     """
     Attempt to locate a fabfile, either explicitly or by searching parent dirs.
 
     Usage docs are in docs/usage/fabfiles.rst, in "Fabfile discovery."
     """
-    if os.path.dirname(state.env.fabfile):
-        expanded = os.path.expanduser(state.env.fabfile)
-        if os.path.exists(expanded):
-            return os.path.abspath(expanded)
-        else:
-            return None
+    # Obtain env value
+    names = [state.env.fabfile]
+    # Create .py version if necessary
+    if not names[0].endswith('.py'):
+        names += [names[0] + '.py']
+    # Does the name contain path elements?
+    if os.path.dirname(names[0]):
+        # If so, expand home-directory markers and test for existence
+        for name in names:
+            expanded = os.path.expanduser(name)
+            if os.path.exists(expanded):
+                if name.endswith('.py') or _is_package(expanded):
+                    return os.path.abspath(expanded)
     else:
+        # Otherwise, start in cwd and work downwards towards filesystem root
         path = '.'
         # Stop before falling off root of filesystem (should be platform
         # agnostic)
         while os.path.split(os.path.abspath(path))[1]:
-            joined = os.path.join(path, state.env.fabfile)
-            if os.path.exists(joined):
-                return os.path.abspath(joined)
+            for name in names:
+                joined = os.path.join(path, name)
+                if os.path.exists(joined):
+                    if name.endswith('.py') or _is_package(joined):
+                        return os.path.abspath(joined)
             path = os.path.join('..', path)
-        return None
+    # Implicit 'return None' if nothing was found
 
 
 def is_task(tup):
@@ -81,7 +101,7 @@ def is_task(tup):
     )
 
 
-def load_fabfile(path):
+def load_fabfile(path, importer=None):
     """
     Import given fabfile path and return (docstring, callables).
 
@@ -89,6 +109,8 @@ def load_fabfile(path):
     dictionary of ``{'name': callable}`` containing all callables which pass
     the "is a Fabric task" test.
     """
+    if importer is None:
+        importer = __import__
     # Get directory and fabfile name
     directory, fabfile = os.path.split(path)
     # If the directory isn't in the PYTHONPATH, add it so our import will work
@@ -109,7 +131,7 @@ def load_fabfile(path):
             sys.path.insert(0, directory)
             del sys.path[i + 1]
     # Perform the import (trimming off the .py)
-    imported = __import__(os.path.splitext(fabfile)[0])
+    imported = importer(os.path.splitext(fabfile)[0])
     # Remove directory from path if we added it ourselves (just to be neat)
     if added_to_path:
         del sys.path[0]
@@ -156,6 +178,14 @@ def parse_options():
         help="print list of possible commands and exit"
     )
 
+    # Like --list, but text processing friendly
+    parser.add_option('--shortlist',
+        action='store_true',
+        dest='shortlist',
+        default=False,
+        help="print non-verbose list of possible commands and exit"
+    )
+
     # Display info about a specific command
     parser.add_option('-d', '--display',
         metavar='COMMAND',
@@ -178,6 +208,10 @@ def parse_options():
     return parser, opts, args
 
 
+def _command_names():
+    return sorted(commands.keys())
+
+
 def list_commands(docstring):
     """
     Print all found commands/tasks, then exit. Invoked with ``-l/--list.``
@@ -192,8 +226,7 @@ def list_commands(docstring):
     max_len = reduce(lambda a, b: max(a, len(b)), commands.keys(), 0)
     sep = '  '
     trail = '...'
-    names = sorted(commands.keys())
-    for name in names:
+    for name in _command_names():
         output = None
         # Print first line of docstring
         func = commands[name]
@@ -212,6 +245,14 @@ def list_commands(docstring):
     sys.exit(0)
 
 
+def shortlist():
+    """
+    Print all task names separated by newlines with no embellishment.
+    """
+    print("\n".join(_command_names()))
+    sys.exit(0)
+
+
 def display_command(command):
     """
     Print command function's docstring, then exit. Invoked with -d/--display.
@@ -232,6 +273,33 @@ def display_command(command):
     sys.exit(0)
 
 
+def _escape_split(sep, argstr):
+    """
+    Allows for escaping of the separator: e.g. task:arg='foo\, bar'
+
+    It should be noted that the way bash et. al. do command line parsing, those
+    single quotes are required.
+    """
+    escaped_sep = r'\%s' % sep
+
+    if escaped_sep not in argstr:
+        return argstr.split(sep)
+
+    before, _, after = argstr.partition(escaped_sep)
+    startlist = before.split(sep) # a regular split is fine here
+    unfinished = startlist[-1]
+    startlist = startlist[:-1]
+
+    # recurse because there may be more escaped separators
+    endlist = _escape_split(sep, after)
+
+    # finish building the escaped value. we use endlist[0] becaue the first
+    # part of the string sent in recursion is the rest of the escaped value.
+    unfinished += sep + endlist[0]
+
+    return startlist + [unfinished] + endlist[1:] # put together all the parts
+
+
 def parse_arguments(arguments):
     """
     Parse string list into list of tuples: command, args, kwargs, hosts, roles.
@@ -246,9 +314,9 @@ def parse_arguments(arguments):
         roles = []
         if ':' in cmd:
             cmd, argstr = cmd.split(':', 1)
-            for pair in argstr.split(','):
+            for pair in _escape_split(',', argstr):
                 k, _, v = pair.partition('=')
-                if v:
+                if _:
                     # Catch, interpret host/hosts/role/roles kwargs
                     if k in ['host', 'hosts', 'role', 'roles']:
                         if k == 'host':
@@ -268,6 +336,13 @@ def parse_arguments(arguments):
     return cmds
 
 
+def parse_remainder(arguments):
+    """
+    Merge list of "remainder arguments" into a single command string.
+    """
+    return ' '.join(arguments)
+
+
 def _merge(hosts, roles):
     """
     Merge given host and role lists into one list of deduped hosts.
@@ -280,13 +355,22 @@ def _merge(hosts, roles):
         ))
 
     # Look up roles, turn into flat list of hosts
-    role_hosts = (
-        roles
-        and reduce(add, [state.env.roledefs[y] for y in roles])
-        or []
-    )
+    role_hosts = []
+    for role in roles:
+        value = state.env.roledefs[role]
+        # Handle "lazy" roles (callables)
+        if callable(value):
+            value = value()
+        role_hosts += value
     # Return deduped combo of hosts and role_hosts
-    return list(set(hosts + role_hosts))
+    return list(set(_clean_hosts(hosts + role_hosts)))
+
+
+def _clean_hosts(host_list):
+    """
+    Clean host strings to ensure no trailing whitespace, etc.
+    """
+    return [host.strip() for host in host_list]
 
 
 def get_hosts(command, cli_hosts, cli_roles):
@@ -336,6 +420,10 @@ def main():
         # Parse command line options
         parser, options, arguments = parse_options()
 
+        # Handle regular args vs -- args
+        arguments = parser.largs
+        remainder_arguments = parser.rargs
+
         # Update env with any overridden option values
         # NOTE: This needs to remain the first thing that occurs
         # post-parsing, since so many things hinge on the values in env.
@@ -357,7 +445,9 @@ def main():
 
         # Handle case where we were called bare, i.e. just "fab", and print
         # a help message.
-        if not (options.list_commands or options.display or arguments):
+        actions = (options.list_commands, options.shortlist, options.display,
+            arguments, remainder_arguments)
+        if not any(actions):
             parser.print_help()
             sys.exit(1)
 
@@ -366,7 +456,7 @@ def main():
 
         # Find local fabfile path or abort
         fabfile = find_fabfile()
-        if not fabfile:
+        if not fabfile and not remainder_arguments:
             abort("Couldn't find any fabfiles!")
 
         # Store absolute path to fabfile in case anyone needs it
@@ -375,16 +465,24 @@ def main():
         # Load fabfile (which calls its module-level code, including
         # tweaks to env values) and put its commands in the shared commands
         # dict
-        docstring, callables = load_fabfile(fabfile)
-        commands.update(callables)
+        if fabfile:
+            docstring, callables = load_fabfile(fabfile)
+            commands.update(callables)
 
         # Abort if no commands found
-        if not commands:
+        if not commands and not remainder_arguments:
             abort("Fabfile didn't contain any commands!")
 
         # Now that we're settled on a fabfile, inform user.
         if state.output.debug:
-            print("Using fabfile '%s'" % fabfile)
+            if fabfile:
+                print("Using fabfile '%s'" % fabfile)
+            else:
+                print("No fabfile loaded -- remainder command only")
+
+        # Non-verbose command list
+        if options.shortlist:
+            shortlist()
 
         # Handle list-commands option (now that commands are loaded)
         if options.list_commands:
@@ -395,14 +493,17 @@ def main():
             display_command(options.display)
 
         # If user didn't specify any commands to run, show help
-        if not arguments:
+        if not (arguments or remainder_arguments):
             parser.print_help()
             sys.exit(0) # Or should it exit with error (1)?
 
         # Parse arguments into commands to run (plus args/kwargs/hosts)
         commands_to_run = parse_arguments(arguments)
 
-        # Figure out if any specified names are invalid
+        # Parse remainders into a faux "command" to execute
+        remainder_command = parse_remainder(remainder_arguments)
+
+        # Figure out if any specified task names are invalid
         unknown_commands = []
         for tup in commands_to_run:
             if tup[0] not in commands:
@@ -413,6 +514,16 @@ def main():
             abort("Command(s) not found:\n%s" \
                 % indent(unknown_commands))
 
+        # Generate remainder command and insert into commands, commands_to_run
+        if remainder_command:
+            r = '<remainder>'
+            commands[r] = lambda: api.run(remainder_command)
+            commands_to_run.append((r, [], {}, [], []))
+
+        if state.output.debug:
+            names = ", ".join(x[0] for x in commands_to_run)
+            print("Commands to run: %s" % names)
+
         # At this point all commands must exist, so execute them in order.
         for name, args, kwargs, cli_hosts, cli_roles in commands_to_run:
             # Get callable by itself
@@ -424,13 +535,13 @@ def main():
                 command, cli_hosts, cli_roles)
             # If hosts found, execute the function on each host in turn
             for host in hosts:
-                username, hostname, port = normalize(host)
-                state.env.host_string = host
-                state.env.host = hostname
                 # Preserve user
                 prev_user = state.env.user
-                state.env.user = username
-                state.env.port = port
+                # Split host string and apply to env dict
+                username, hostname, port = interpret_host_string(host)
+                # Log to stdout
+                if state.output.running:
+                    print("[%s] Executing task '%s'" % (host, name))
                 # Actually run command
                 commands[name](*args, **kwargs)
                 # Put old user back
@@ -453,11 +564,5 @@ def main():
         # we might leave stale threads if we don't explicitly exit()
         sys.exit(1)
     finally:
-        # Explicitly disconnect from all servers
-        for key in connections.keys():
-            if state.output.status:
-                print "Disconnecting from %s..." % denormalize(key),
-            connections[key].close()
-            if state.output.status:
-                print "done."
+        disconnect_all()
     sys.exit(0)
